@@ -1,13 +1,21 @@
-#!/bin/bash
+#!/usr/bin/env bash
 
 set -e
 
+# optional env variable override
+if [ -n "$SYSREPOCFG_EXECUTABLE" ]; then
+    SYSREPOCFG="$SYSREPOCFG_EXECUTABLE"
 # avoid problems with sudo PATH
-if [ `id -u` -eq 0 ]; then
+elif [ `id -u` -eq 0 ]; then
     SYSREPOCFG=`su -c 'which sysrepocfg' -l $USER`
-    OPENSSL=`su -c 'which openssl' -l $USER`
 else
     SYSREPOCFG=`which sysrepocfg`
+fi
+
+# avoid problems with sudo PATH
+if [ `id -u` -eq 0 ]; then
+    OPENSSL=`su -c 'which openssl' -l $USER`
+else
     OPENSSL=`which openssl`
 fi
 
