@@ -56,7 +56,7 @@ np_get_nc_sess_user(sr_session_ctx_t *session)
     struct nc_session *nc_sess = NULL;
     uint32_t nc_sid, i;
 
-    nc_sid = sr_session_get_nc_id(session);
+    nc_sid = sr_session_get_event_nc_id(session);
     for (i = 0; (nc_sess = nc_ps_get_session(np2srv.nc_ps, i)); ++i) {
         if (nc_session_get_id(nc_sess) == nc_sid) {
             break;
@@ -69,6 +69,25 @@ np_get_nc_sess_user(sr_session_ctx_t *session)
     return nc_session_get_username(nc_sess);
 }
 
+sr_session_ctx_t *
+np_get_user_sess(sr_session_ctx_t *ev_sess)
+{
+    struct nc_session *nc_sess = NULL;
+    uint32_t nc_sid, i;
+
+    nc_sid = sr_session_get_event_nc_id(ev_sess);
+    for (i = 0; (nc_sess = nc_ps_get_session(np2srv.nc_ps, i)); ++i) {
+        if (nc_session_get_id(nc_sess) == nc_sid) {
+            break;
+        }
+    }
+    if (!nc_sess) {
+        return NULL;
+    }
+
+    return nc_session_get_data(nc_sess);
+}
+
 void
 np2srv_ntf_new_cb(sr_session_ctx_t *UNUSED(session), const sr_ev_notif_type_t notif_type, const struct lyd_node *notif,
         time_t timestamp, void *private_data)
@@ -461,27 +480,28 @@ strws(const char *str)
 }
 
 static int
-op_filter_xpath_add_filter(char *new_filter, char ***filters, int *filter_count)
+op_filter_xpath_add_filter(const char *new_filter, int selection, struct np2_filter *filter)
 {
-    char **filters_new;
+    void *mem;
 
-    filters_new = realloc(*filters, (*filter_count + 1) * sizeof **filters);
-    if (!filters_new) {
+    mem = realloc(filter->filters, (filter->count + 1) * sizeof *filter->filters);
+    if (!mem) {
         EMEM;
         return -1;
     }
-    ++(*filter_count);
-    *filters = filters_new;
-    (*filters)[*filter_count - 1] = new_filter;
+    filter->filters = mem;
+    filter->filters[filter->count].str = strdup(new_filter);
+    filter->filters[filter->count].selection = selection;
+    ++filter->count;
 
     return 0;
 }
 
 static int
-filter_xpath_buf_add_attrs(struct ly_ctx *ctx, struct lyxml_attr *attr, char **buf, int size)
+filter_xpath_buf_append_attrs(struct ly_ctx *ctx, const struct lyxml_attr *attr, char **buf, int size)
 {
     const struct lys_module *module;
-    struct lyxml_attr *next;
+    const struct lyxml_attr *next;
     int new_size;
     char *buf_new;
 
@@ -512,7 +532,7 @@ filter_xpath_buf_add_attrs(struct ly_ctx *ctx, struct lyxml_attr *attr, char **b
 }
 
 static char *
-filter_xpath_buf_get_content(struct ly_ctx *ctx, struct lyxml_elem *elem)
+filter_xpath_buf_get_content(struct ly_ctx *ctx, const struct lyxml_elem *elem)
 {
     const char *start;
     size_t len;
@@ -525,7 +545,7 @@ filter_xpath_buf_get_content(struct ly_ctx *ctx, struct lyxml_elem *elem)
     start = lydict_insert(ctx, start, len);
 
     ly_log_options(0);
-    ret = ly_path_xml2json(ctx, start, elem);
+    ret = ly_path_xml2json(ctx, start, (struct lyxml_elem *)elem);
     ly_log_options(LY_LOLOG | LY_LOSTORE_LAST);
 
     if (!ret) {
@@ -538,8 +558,8 @@ filter_xpath_buf_get_content(struct ly_ctx *ctx, struct lyxml_elem *elem)
 
 /* top-level content node with optional namespace and attributes */
 static int
-filter_xpath_buf_add_top_content(struct ly_ctx *ctx, struct lyxml_elem *elem, const char *elem_module_name,
-                                char ***filters, int *filter_count)
+filter_xpath_buf_add_top_content(struct ly_ctx *ctx, const struct lyxml_elem *elem, const char *elem_module_name,
+                                 struct np2_filter *filter)
 {
     int size;
     char *buf, *content;
@@ -547,7 +567,7 @@ filter_xpath_buf_add_top_content(struct ly_ctx *ctx, struct lyxml_elem *elem, co
     content = filter_xpath_buf_get_content(ctx, elem);
 
     size = 1 + strlen(elem_module_name) + 1 + strlen(elem->name) + 9 + strlen(content) + 3;
-    buf = malloc(size * sizeof(char));
+    buf = malloc(size);
     if (!buf) {
         EMEM;
         free(content);
@@ -556,7 +576,7 @@ filter_xpath_buf_add_top_content(struct ly_ctx *ctx, struct lyxml_elem *elem, co
     sprintf(buf, "/%s:%s[text()='%s']", elem_module_name, elem->name, content);
     free(content);
 
-    size = filter_xpath_buf_add_attrs(ctx, elem->attr, &buf, size);
+    size = filter_xpath_buf_append_attrs(ctx, elem->attr, &buf, size);
     if (!size) {
         free(buf);
         return 0;
@@ -565,18 +585,19 @@ filter_xpath_buf_add_top_content(struct ly_ctx *ctx, struct lyxml_elem *elem, co
         return -1;
     }
 
-    if (op_filter_xpath_add_filter(buf, filters, filter_count)) {
+    if (op_filter_xpath_add_filter(buf, 0, filter)) {
         free(buf);
         return -1;
     }
 
+    free(buf);
     return 0;
 }
 
 /* content node with optional namespace and attributes */
 static int
-filter_xpath_buf_add_content(struct ly_ctx *ctx, struct lyxml_elem *elem, const char *elem_module_name,
-                            const char *last_ns, char **buf, int size)
+filter_xpath_buf_append_content(struct ly_ctx *ctx, const struct lyxml_elem *elem, const char *elem_module_name,
+                                const char *last_ns, char **buf, int size)
 {
     const struct lys_module *module;
     int new_size;
@@ -594,7 +615,7 @@ filter_xpath_buf_add_content(struct ly_ctx *ctx, struct lyxml_elem *elem, const
     }
 
     new_size = size + 1 + (elem_module_name ? strlen(elem_module_name) + 1 : 0) + strlen(elem->name);
-    buf_new = realloc(*buf, new_size * sizeof(char));
+    buf_new = realloc(*buf, new_size);
     if (!buf_new) {
         EMEM;
         return -1;
@@ -604,17 +625,15 @@ filter_xpath_buf_add_content(struct ly_ctx *ctx, struct lyxml_elem *elem, const
             elem->name);
     size = new_size;
 
-    size = filter_xpath_buf_add_attrs(ctx, elem->attr, buf, size);
-    if (!size) {
-        return 0;
-    } else if (size < 1) {
-        return -1;
+    size = filter_xpath_buf_append_attrs(ctx, elem->attr, buf, size);
+    if (size < 1) {
+        return size;
     }
 
     content = filter_xpath_buf_get_content(ctx, elem);
 
     new_size = size + 2 + strlen(content) + 2;
-    buf_new = realloc(*buf, new_size * sizeof(char));
+    buf_new = realloc(*buf, new_size);
     if (!buf_new) {
         EMEM;
         free(content);
@@ -635,8 +654,8 @@ filter_xpath_buf_add_content(struct ly_ctx *ctx, struct lyxml_elem *elem, const
 
 /* containment/selection node with optional namespace and attributes */
 static int
-filter_xpath_buf_add_node(struct ly_ctx *ctx, struct lyxml_elem *elem, const char *elem_module_name,
-                         const char *last_ns, char **buf, int size)
+filter_xpath_buf_append_node(struct ly_ctx *ctx, const struct lyxml_elem *elem, const char *elem_module_name,
+                             const char *last_ns, char **buf, int size)
 {
     const struct lys_module *module;
     int new_size;
@@ -654,7 +673,7 @@ filter_xpath_buf_add_node(struct ly_ctx *ctx, struct lyxml_elem *elem, const cha
     }
 
     new_size = size + 1 + (elem_module_name ? strlen(elem_module_name) + 1 : 0) + strlen(elem->name);
-    buf_new = realloc(*buf, new_size * sizeof(char));
+    buf_new = realloc(*buf, new_size);
     if (!buf_new) {
         EMEM;
         return -1;
@@ -664,121 +683,100 @@ filter_xpath_buf_add_node(struct ly_ctx *ctx, struct lyxml_elem *elem, const cha
             elem->name);
     size = new_size;
 
-    size = filter_xpath_buf_add_attrs(ctx, elem->attr, buf, size);
-
+    size = filter_xpath_buf_append_attrs(ctx, elem->attr, buf, size);
     return size;
 }
 
-/* buf is spent in the function, removes content match nodes from elem->child list! */
 static int
-filter_xpath_buf_add(struct ly_ctx *ctx, struct lyxml_elem *elem, const char *elem_module_name, const char *last_ns,
-                    char **buf, int size, char ***filters, int *filter_count)
+filter_xpath_buf_add_r(struct ly_ctx *ctx, const struct lyxml_elem *elem, const char *elem_module_name,
+                     const char *last_ns, char **buf, int size, struct np2_filter *filter)
 {
-    struct lyxml_elem *temp, *child;
-    int new_size;
-    char *buf_new;
-    int only_content_match_node = 1;
+    struct lyxml_elem *child;
+    int s, only_content_match;
 
-    /* containment node, selection node */
-    size = filter_xpath_buf_add_node(ctx, elem, elem_module_name, last_ns, buf, size);
-    if (!size) {
-        free(*buf);
-        *buf = NULL;
-        return 0;
-    } else if (size < 1) {
-        goto error;
+    /* containment node or selection node */
+    size = filter_xpath_buf_append_node(ctx, elem, elem_module_name, last_ns, buf, size);
+    if (size < 1) {
+        return size;
     }
 
-    /* content match node */
-    LY_TREE_FOR_SAFE(elem->child, temp, child) {
-        if (!child->child && child->content && !strws(child->content)) {
-            size = filter_xpath_buf_add_content(ctx, child, elem_module_name, last_ns, buf, size);
-            if (!size) {
-                free(*buf);
-                *buf = NULL;
-                return 0;
-            } else if (size < 1) {
-                goto error;
-            }
-        } else {
-            only_content_match_node = 0;
+    if (!elem->child) {
+        /* just a selection node */
+        if (op_filter_xpath_add_filter(*buf, 1, filter)) {
+            return -1;
         }
-    }
-
-    /* that is it, it seems */
-    if (only_content_match_node) {
-        if (op_filter_xpath_add_filter(*buf, filters, filter_count)) {
-            goto error;
-        }
-        *buf = NULL;
         return 0;
     }
 
-    /* that is it for this filter depth, now we branch with every new node except last */
+    /* append child content match nodes */
+    only_content_match = 1;
     LY_TREE_FOR(elem->child, child) {
-        if (!child->next) {
-            buf_new = *buf;
-            *buf = NULL;
-        } else {
-            buf_new = malloc(size * sizeof(char));
-            if (!buf_new) {
-                EMEM;
-                goto error;
+        if (!child->child && child->content && !strws(child->content)) {
+            /* there is a content filter, append all of them */
+            size = filter_xpath_buf_append_content(ctx, child, elem_module_name, last_ns, buf, size);
+            if (size < 1) {
+                return size;
             }
-            memcpy(buf_new, *buf, size * sizeof(char));
+        } else {
+            only_content_match = 0;
         }
-        new_size = size;
+    }
 
-        /* child containment node */
+    if (only_content_match) {
+        /* there are only content match nodes so we retrieve this filter as a subtree */
+        if (op_filter_xpath_add_filter(*buf, 0, filter)) {
+            return -1;
+        }
+
+        return 0;
+    }
+    /* else there are some other filters so the current filter just restricts all the nested ones, is not retrieved
+     * as a standalone subtree */
+
+    /* that is it for this filter depth, now we branch with every new node */
+    LY_TREE_FOR(elem->child, child) {
         if (child->child) {
-            filter_xpath_buf_add(ctx, child, NULL, last_ns, &buf_new, new_size, filters, filter_count);
-
-        /* child selection node or content match node */
-        } else {
-            new_size = filter_xpath_buf_add_node(ctx, child, NULL, last_ns, &buf_new, new_size);
-            if (!new_size) {
-                free(buf_new);
+            /* child containment node */
+            filter_xpath_buf_add_r(ctx, child, NULL, last_ns, buf, size, filter);
+        } else if (!child->content || strws(child->content)) {
+            /* child selection node */
+            s = filter_xpath_buf_append_node(ctx, child, NULL, last_ns, buf, size);
+            if (!s) {
                 continue;
-            } else if (new_size < 1) {
-                free(buf_new);
-                goto error;
+            } else if (s < 0) {
+                return s;
             }
 
-            if (op_filter_xpath_add_filter(buf_new, filters, filter_count)) {
-                goto error;
+            if (op_filter_xpath_add_filter(*buf, 1, filter)) {
+                return -1;
             }
-        }
+        } /* else child content node, already handled */
     }
 
     return 0;
-
-error:
-    free(*buf);
-    return -1;
 }
 
-/* modifies elem XML tree! */
 static int
-op_filter_build_xpath_from_subtree(struct ly_ctx *ctx, struct lyxml_elem *elem, char ***filters, int *filter_count)
+op_filter_build_xpath_from_subtree(struct ly_ctx *ctx, const struct lyxml_elem *elem, struct np2_filter *filter)
 {
     const struct lys_module *module, **modules, **modules_new;
     const struct lys_node *node;
-    struct lyxml_elem *next;
-    char *buf;
+    const struct lyxml_elem *iter;
+    char *buf = NULL;
     uint32_t i, module_count;
 
-    LY_TREE_FOR(elem, next) {
+    LY_TREE_FOR(elem, iter) {
         /* first filter node, it must always have a namespace */
         modules = NULL;
         module_count = 0;
-        if (next->ns && strcmp(next->ns->value, "urn:ietf:params:xml:ns:netconf:base:1.0")) {
+        if (iter->ns && strcmp(iter->ns->value, "urn:ietf:params:xml:ns:netconf:base:1.0")) {
             modules = malloc(sizeof *modules);
             if (!modules) {
                 EMEM;
                 goto error;
             }
             module_count = 1;
-            modules[0] = ly_ctx_get_module_by_ns(ctx, next->ns->value, NULL, 1);
+            modules[0] = ly_ctx_get_module_by_ns(ctx, iter->ns->value, NULL, 1);
             if (!modules[0]) {
                 /* not really an error */
                 free(modules);
@@ -789,7 +787,7 @@ op_filter_build_xpath_from_subtree(struct ly_ctx *ctx, struct lyxml_elem *elem,
             while ((module = ly_ctx_get_module_iter(ctx, &i))) {
                 node = NULL;
                 while ((node = lys_getnext(node, NULL, module, 0))) {
-                    if (!strcmp(node->name, next->name)) {
+                    if (!strcmp(node->name, iter->name)) {
                         modules_new = realloc(modules, (module_count + 1) * sizeof *modules);
                         if (!modules_new) {
                             EMEM;
@@ -804,16 +802,15 @@ op_filter_build_xpath_from_subtree(struct ly_ctx *ctx, struct lyxml_elem *elem,
             }
         }
 
-        buf = NULL;
         for (i = 0; i < module_count; ++i) {
-            if (!next->child && next->content && !strws(next->content)) {
+            if (!iter->child && iter->content && !strws(iter->content)) {
                 /* special case of top-level content match node */
-                if (filter_xpath_buf_add_top_content(ctx, next, modules[i]->name, filters, filter_count)) {
+                if (filter_xpath_buf_add_top_content(ctx, iter, modules[i]->name, filter)) {
                     goto error;
                 }
             } else {
                 /* containment or selection node */
-                if (filter_xpath_buf_add(ctx, next, modules[i]->name, modules[i]->ns, &buf, 1, filters, filter_count)) {
+                if (filter_xpath_buf_add_r(ctx, iter, modules[i]->name, modules[i]->ns, &buf, 1, filter)) {
                     goto error;
                 }
             }
@@ -821,25 +818,36 @@ op_filter_build_xpath_from_subtree(struct ly_ctx *ctx, struct lyxml_elem *elem,
         free(modules);
     }
 
+    free(buf);
     return 0;
 
 error:
+    free(buf);
     free(modules);
-    for (i = 0; (signed)i < *filter_count; ++i) {
-        free((*filters)[i]);
-    }
-    free(*filters);
+    op_filter_erase(filter);
     return -1;
 }
 
+void
+op_filter_erase(struct np2_filter *filter)
+{
+    int i;
+
+    for (i = 0; i < filter->count; ++i) {
+        free(filter->filters[i].str);
+    }
+    free(filter->filters);
+    filter->filters = NULL;
+    filter->count = 0;
+}
+
 int
-op_filter_create(struct lyd_node *filter_node, char ***filters, int *filter_count)
+op_filter_create(struct lyd_node *filter_node, struct np2_filter *filter)
 {
     struct lyd_attr *attr;
     struct lyxml_elem *subtree_filter;
     struct ly_ctx *ly_ctx;
     int free_filter, ret;
-    char *path;
 
     ly_ctx = lyd_node_module(filter_node)->ctx;
 
@@ -890,9 +898,9 @@ op_filter_create(struct lyd_node *filter_node, char ***filters, int *filter_coun
             return -1;
         }
 
-        ret = op_filter_build_xpath_from_subtree(ly_ctx, subtree_filter, filters, filter_count);
+        ret = op_filter_build_xpath_from_subtree(ly_ctx, subtree_filter, filter);
         if (free_filter) {
-            lyxml_free(ly_ctx, subtree_filter);
+            lyxml_free_withsiblings(ly_ctx, subtree_filter);
         }
         if (ret) {
             return -1;
@@ -903,16 +911,105 @@ op_filter_create(struct lyd_node *filter_node, char ***filters, int *filter_coun
             /* empty select, okay, I guess... */
             return 0;
         }
-        path = strdup(attr->value_str);
-        if (!path) {
-            EMEM;
-            return -1;
-        }
-        if (op_filter_xpath_add_filter(path, filters, filter_count)) {
-            free(path);
+        if (op_filter_xpath_add_filter(attr->value_str, 1, filter)) {
             return -1;
         }
     }
 
     return 0;
 }
+
+int
+op_filter_data_get(sr_session_ctx_t *session, uint32_t max_depth, sr_get_oper_options_t get_opts,
+        const struct np2_filter *filter, sr_session_ctx_t *ev_sess, struct lyd_node **data)
+{
+    const sr_error_info_t *err_info;
+    struct lyd_node *node;
+    int i, rc;
+
+    for (i = 0; i < filter->count; ++i) {
+        /* get the selected data */
+        rc = sr_get_data(session, filter->filters[i].str, max_depth, np2srv.sr_timeout, get_opts, &node);
+        if (rc) {
+            ERR("Getting data \"%s\" from sysrepo failed (%s).", filter->filters[i].str, sr_strerror(rc));
+            sr_get_error(session, &err_info);
+            sr_set_error(ev_sess, err_info->err[0].xpath, err_info->err[0].message);
+            return rc;
+        }
+
+        /* merge */
+        if (!*data) {
+            *data = node;
+        } else if (node && lyd_merge(*data, node, LYD_OPT_DESTRUCT | LYD_OPT_EXPLICIT)) {
+            lyd_free_withsiblings(node);
+            return SR_ERR_LY;
+        }
+    }
+
+    return SR_ERR_OK;
+}
+
+int
+op_filter_data_filter(struct lyd_node **data, const struct np2_filter *filter, int with_selection,
+        struct lyd_node **filtered_data)
+{
+    struct lyd_node *node;
+    int i, has_filter = 0, rc = SR_ERR_OK;
+    struct ly_set *set = NULL;
+    uint32_t j;
+
+    if (!*data) {
+        /* nothing to filter */
+        return SR_ERR_OK;
+    }
+
+    for (i = 0; i < filter->count; i++) {
+        if (!with_selection && filter->filters[i].selection) {
+            continue;
+        }
+        has_filter = 1;
+
+        /* apply content (or even selection) filter */
+        set = lyd_find_path(*data, filter->filters[i].str);
+        if (!set) {
+            rc = SR_ERR_LY;
+            goto cleanup;
+        }
+
+        for (j = 0; j < set->number; ++j) {
+            node = lyd_dup(set->set.d[j], LYD_DUP_OPT_RECURSIVE | LYD_DUP_OPT_WITH_PARENTS | LYD_DUP_OPT_WITH_KEYS |
+                    LYD_DUP_OPT_WITH_WHEN);
+            if (!node) {
+                rc = SR_ERR_LY;
+                goto cleanup;
+            }
+
+            /* always find parent */
+            while (node->parent) {
+                node = node->parent;
+            }
+
+            /* merge */
+            if (!*filtered_data) {
+                *filtered_data = node;
+            } else if (node && lyd_merge(*filtered_data, node, LYD_OPT_DESTRUCT | LYD_OPT_EXPLICIT)) {
+                lyd_free_withsiblings(node);
+                rc = SR_ERR_LY;
+                goto cleanup;
+            }
+        }
+
+        ly_set_free(set);
+        set = NULL;
+    }
+
+    if (!has_filter) {
+        /* no filter, just use all the data */
+        *filtered_data = *data;
+        *data = NULL;
+    }
+
+cleanup:
+    ly_set_free(set);
+    return rc;
+}
