@@ -367,6 +367,10 @@ err_access_denied:
             e = nc_err(NC_ERR_INVALID_VALUE, NC_ERR_TYPE_PROT);
             nc_err_set_msg(e, message, "en");
             break;
+        } else if (strstr(message, "already subscribed")) {
+            e = nc_err(NC_ERR_OP_FAILED, NC_ERR_TYPE_PROT);
+            nc_err_set_msg(e, message, "en");
+            break;
         }
         e = nc_err(NC_ERR_OP_FAILED, NC_ERR_TYPE_APP);
         nc_err_set_msg(e, message, "en");
@@ -506,17 +510,17 @@ np2srv_diff_check_cb(sr_session_ctx_t *session, const struct lyd_node *diff)
     char *path;
     const char *user;
 
-    if (ATOMIC_LOAD_RELAXED(skip_nacm_sr_sid) == sr_session_get_id(session)) {
-        /* skip the NACM check */
-        return SR_ERR_OK;
-    }
-
     user = np_get_nc_sess_user(session);
     if (!user) {
         EINT;
         return SR_ERR_INTERNAL;
     }
 
+    if (ATOMIC_LOAD_RELAXED(skip_nacm_sr_sid) == sr_session_get_event_sr_id(session)) {
+        /* skip the NACM check */
+        return SR_ERR_OK;
+    }
+
     if ((node = ncac_check_diff(diff, user))) {
         /* access denied */
         path = lys_data_path(node->schema);
@@ -741,7 +745,6 @@ error:
 static int
 server_rpc_subscribe(void)
 {
-    const char *xpath;
     int rc;
 
 #define SR_RPC_SUBSCR(xpath, cb) \
@@ -769,14 +772,8 @@ server_rpc_subscribe(void)
     SR_RPC_SUBSCR("/ietf-netconf:discard-changes", np2srv_rpc_discard_cb);
     SR_RPC_SUBSCR("/ietf-netconf:validate", np2srv_rpc_validate_cb);
 
-    xpath = "/notifications:create-subscription";
-    /* subscribes to notifications, needs special flag */
-    rc = sr_rpc_subscribe_tree(np2srv.sr_sess, xpath, np2srv_rpc_subscribe_cb, NULL, 0, SR_SUBSCR_UNLOCKED | SR_SUBSCR_CTX_REUSE,
-            &np2srv.sr_rpc_sub);
-    if (rc != SR_ERR_OK) {
-        ERR("Subscribing for \"%s\" RPC failed (%s).", xpath, sr_strerror(rc));
-        goto error;
-    }
+    /* subscribe to create-subscription */
+    SR_RPC_SUBSCR("/notifications:create-subscription", np2srv_rpc_subscribe_cb);
 
     /* subscribe to NMDA RPCs */
     SR_RPC_SUBSCR("/ietf-netconf-nmda:get-data", np2srv_rpc_getdata_cb);
@@ -982,8 +979,7 @@ worker_thread(void *arg)
 
     while (ATOMIC_LOAD_RELAXED(loop_continue)) {
         /* try to accept new NETCONF sessions */
-        if (nc_server_endpt_count()
-                && (!np2srv.nc_max_sessions || (nc_ps_session_count(np2srv.nc_ps) < np2srv.nc_max_sessions))) {
+        if (nc_server_endpt_count()) {
             msgtype = nc_accept(0, &ncs);
             if (msgtype == NC_MSG_HELLO) {
                 np2srv_new_session_cb(NULL, ncs);
