@@ -62,24 +62,23 @@ np2srv_get_first_ns(const char *expr, const char **start, int *len)
  * @brief Get generic filters in the form of "/module:*" from exact xpath filters.
  */
 static int
-np2srv_get_rpc_module_filters(char **filters, int filter_count, char ***mod_filters, int *mod_filter_count)
+np2srv_get_rpc_module_filters(const struct np2_filter *filter, struct np2_filter *mod_filter)
 {
-    int i, j, len;
+    int i, j, len, selection;
     const char *start;
     char *str;
 
-    *mod_filters = NULL;
-    *mod_filter_count = 0;
-
-    for (i = 0; i < filter_count; ++i) {
-        if (np2srv_get_first_ns(filters[i], &start, &len)) {
+    for (i = 0; i < filter->count; ++i) {
+        if (np2srv_get_first_ns(filter->filters[i].str, &start, &len)) {
             /* not the simple format, use it as it is */
-            str = strdup(filters[i]);
+            str = strdup(filter->filters[i].str);
+            selection = filter->filters[i].selection;
         } else {
             /* get all the data of a module */
             if (asprintf(&str, "/%.*s:*", len, start) == -1) {
                 str = NULL;
             }
+            selection = 1;
         }
 
         if (!str) {
@@ -88,20 +87,21 @@ np2srv_get_rpc_module_filters(char **filters, int filter_count, char ***mod_filt
         }
 
         /* check for a duplicity */
-        for (j = 0; j < *mod_filter_count; ++j) {
-            if (!strcmp(str, (*mod_filters)[j])) {
+        for (j = 0; j < mod_filter->count; ++j) {
+            if (!strcmp(str, mod_filter->filters[j].str)) {
                 break;
             }
         }
-        if (j < *mod_filter_count) {
+        if (j < mod_filter->count) {
             free(str);
             continue;
         }
 
         /* add a new module filter */
-        ++(*mod_filter_count);
-        *mod_filters = realloc(*mod_filters, *mod_filter_count * sizeof **mod_filters);
-        (*mod_filters)[*mod_filter_count - 1] = str;
+        mod_filter->filters = realloc(mod_filter->filters, (mod_filter->count + 1) * sizeof *mod_filter->filters);
+        mod_filter->filters[mod_filter->count].str = str;
+        mod_filter->filters[mod_filter->count].selection = selection;
+        ++mod_filter->count;
     }
 
     return SR_ERR_OK;
@@ -111,20 +111,19 @@ np2srv_get_rpc_module_filters(char **filters, int filter_count, char ***mod_filt
  * @brief Get data for a get RPC.
  */
 static int
-np2srv_get_rpc_data(sr_session_ctx_t *session, char **filters, int filter_count, struct lyd_node **data)
+np2srv_get_rpc_data(sr_session_ctx_t *session, const struct np2_filter *filter, sr_session_ctx_t *ev_sess,
+        struct lyd_node **data)
 {
-    struct lyd_node *node, *ds_data = NULL;
+    struct lyd_node *all_data = NULL;
     sr_datastore_t ds;
     sr_get_oper_options_t get_opts = 0;
-    const sr_error_info_t *err_info;
-    char **mod_filters = NULL;
-    int mod_filter_count = 0, i, rc = SR_ERR_OK;
-    uint32_t j;
+    struct np2_filter mod_filter = {0};
+    int rc = SR_ERR_OK;
     struct ly_set *set = NULL;
 
     /* get generic filters to allow retrieving all possibly needed data first, which are then filtered again
      * (once we have merged config and state data) */
-    rc = np2srv_get_rpc_module_filters(filters, filter_count, &mod_filters, &mod_filter_count);
+    rc = np2srv_get_rpc_module_filters(filter, &mod_filter);
     if (rc) {
         goto cleanup;
     }
@@ -135,22 +134,8 @@ np2srv_get_rpc_data(sr_session_ctx_t *session, char **filters, int filter_count,
 get_sr_data:
     sr_session_switch_ds(session, ds);
 
-    for (i = 0; i < mod_filter_count; ++i) {
-        rc = sr_get_data(session, mod_filters[i], 0, np2srv.sr_timeout, get_opts, &node);
-        if (rc) {
-            ERR("Getting data \"%s\" from sysrepo failed (%s).", mod_filters[i], sr_strerror(rc));
-            sr_get_error(session, &err_info);
-            sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
-            goto cleanup;
-        }
-
-        if (!ds_data) {
-            ds_data = node;
-        } else if (node && lyd_merge(ds_data, node, LYD_OPT_DESTRUCT | LYD_OPT_EXPLICIT)) {
-            lyd_free_withsiblings(node);
-            rc = SR_ERR_LY;
-            goto cleanup;
-        }
+    if ((rc = op_filter_data_get(session, 0, get_opts, &mod_filter, ev_sess, &all_data))) {
+        goto cleanup;
     }
 
     if (ds == SR_DS_RUNNING) {
@@ -160,53 +145,15 @@ get_sr_data:
         goto get_sr_data;
     }
 
-    if (!ds_data) {
-        /* no data to filter */
-        goto cleanup;
-    }
-
     /* now filter only the requested data from the created running data + state data */
-    for (i = 0; i < filter_count; ++i) {
-        set = lyd_find_path(ds_data, filters[i]);
-        if (!set) {
-            rc = SR_ERR_LY;
-            goto cleanup;
-        }
-
-        for (j = 0; j < set->number; ++j) {
-            node = lyd_dup(set->set.d[j], LYD_DUP_OPT_RECURSIVE | LYD_DUP_OPT_WITH_PARENTS | LYD_DUP_OPT_WITH_KEYS |
-                    LYD_DUP_OPT_WITH_WHEN);
-            if (!node) {
-                rc = SR_ERR_LY;
-                goto cleanup;
-            }
-
-            /* always find parent */
-            while (node->parent) {
-                node = node->parent;
-            }
-
-            /* merge */
-            if (!*data) {
-                *data = node;
-            } else if (node && lyd_merge(*data, node, LYD_OPT_DESTRUCT | LYD_OPT_EXPLICIT)) {
-                lyd_free_withsiblings(node);
-                rc = SR_ERR_LY;
-                goto cleanup;
-            }
-        }
-
-        ly_set_free(set);
-        set = NULL;
+    if ((rc = op_filter_data_filter(&all_data, filter, 1, data))) {
+        goto cleanup;
     }
 
 cleanup:
     ly_set_free(set);
-    lyd_free_withsiblings(ds_data);
-    for (i = 0; i < mod_filter_count; ++i) {
-        free(mod_filters[i]);
-    }
-    free(mod_filters);
+    lyd_free_withsiblings(all_data);
+    op_filter_erase(&mod_filter);
     return rc;
 }
 
@@ -214,12 +161,11 @@ cleanup:
  * @brief get data for a get-config RPC.
  */
 static int
-np2srv_getconfig_rpc_data(sr_session_ctx_t *session, char **filters, int filter_count, sr_datastore_t ds,
-        struct lyd_node **data)
+np2srv_getconfig_rpc_data(sr_session_ctx_t *session, const struct np2_filter *filter, sr_datastore_t ds,
+        sr_session_ctx_t *ev_sess, struct lyd_node **data)
 {
-    struct lyd_node *node;
-    const sr_error_info_t *err_info;
-    int i, rc;
+    struct lyd_node *select_data = NULL;
+    int rc = SR_ERR_OK;
 
     /* update sysrepo session datastore */
     sr_session_switch_ds(session, ds);
@@ -227,35 +173,50 @@ np2srv_getconfig_rpc_data(sr_session_ctx_t *session, char **filters, int filter_
     /*
      * create the data tree for the data reply
      */
-    for (i = 0; i < filter_count; i++) {
-        rc = sr_get_data(session, filters[i], 0, np2srv.sr_timeout, 0, &node);
-        if (rc != SR_ERR_OK) {
-            ERR("Getting data \"%s\" from sysrepo failed (%s).", filters[i], sr_strerror(rc));
-            sr_get_error(session, &err_info);
-            sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
-            return rc;
-        }
-
-        if (!*data) {
-            *data = node;
-        } else if (node && lyd_merge(*data, node, LYD_OPT_DESTRUCT | LYD_OPT_EXPLICIT)) {
-            lyd_free_withsiblings(node);
-            return SR_ERR_LY;
-        }
+    if ((rc = op_filter_data_get(session, 0, 0, filter, ev_sess, &select_data))) {
+        goto cleanup;
     }
 
-    return SR_ERR_OK;
+    if ((rc = op_filter_data_filter(&select_data, filter, 0, data))) {
+        goto cleanup;
+    }
+
+cleanup:
+    lyd_free_withsiblings(select_data);
+    return rc;
 }
 
 int
-np2srv_rpc_get_cb(sr_session_ctx_t *session, const char *op_path, const struct lyd_node *input, sr_event_t UNUSED(event),
+np2srv_rpc_get_cb(sr_session_ctx_t *session, const char *op_path, const struct lyd_node *input, sr_event_t event,
         uint32_t UNUSED(request_id), struct lyd_node *output, void *UNUSED(private_data))
 {
     struct lyd_node *node, *data_get = NULL;
-    char **filters = NULL;
-    int filter_count = 0, i, rc = SR_ERR_OK;
+    struct np2_filter filter = {0};
+    int rc = SR_ERR_OK;
+    sr_session_ctx_t *user_sess;
     struct ly_set *nodeset;
     sr_datastore_t ds = 0;
+    char *username = NULL;
+
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
+    /* Get username. It is assumed that right now the NETCONF session cannot end
+     * due to the RPC lock held while np2srv_rpc_cb() is executing (which called this callback).
+     */
+    if ((username = (char *)np_get_nc_sess_user(session))) {
+        if (!(username = strdup(username))) {
+            EMEM;
+            rc = SR_ERR_NOMEM;
+            goto cleanup;
+        }
+    } else {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
 
     /* get know which datastore is being affected for get-config */
     if (!strcmp(op_path, "/ietf-netconf:get-config")) {
@@ -277,22 +238,23 @@ np2srv_rpc_get_cb(sr_session_ctx_t *session, const char *op_path, const struct l
     if (nodeset->number) {
         node = nodeset->set.d[0];
         ly_set_free(nodeset);
-        if (op_filter_create(node, &filters, &filter_count)) {
+        if (op_filter_create(node, &filter)) {
             rc = SR_ERR_INTERNAL;
             goto cleanup;
         }
     } else {
         ly_set_free(nodeset);
 
-        filters = malloc(sizeof *filters);
-        if (!filters) {
+        filter.filters = malloc(sizeof *filter.filters);
+        if (!filter.filters) {
             EMEM;
             rc = SR_ERR_NOMEM;
             goto cleanup;
         }
-        filter_count = 1;
-        filters[0] = strdup("/*");
-        if (!filters[0]) {
+        filter.count = 1;
+        filter.filters[0].str = strdup("/*");
+        filter.filters[0].selection = 1;
+        if (!filter.filters[0].str) {
             EMEM;
             rc = SR_ERR_NOMEM;
             goto cleanup;
@@ -301,18 +263,26 @@ np2srv_rpc_get_cb(sr_session_ctx_t *session, const char *op_path, const struct l
 
     /* we do not care here about with-defaults mode, it does not change anything */
 
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* get filtered data */
     if (!strcmp(op_path, "/ietf-netconf:get-config")) {
-        rc = np2srv_getconfig_rpc_data(session, filters, filter_count, ds, &data_get);
+        rc = np2srv_getconfig_rpc_data(user_sess, &filter, ds, session, &data_get);
     } else {
-        rc = np2srv_get_rpc_data(session, filters, filter_count, &data_get);
+        rc = np2srv_get_rpc_data(user_sess, &filter, session, &data_get);
     }
     if (rc) {
         goto cleanup;
     }
 
     /* perform correct NACM filtering */
-    ncac_check_data_read_filter(&data_get, np_get_nc_sess_user(session));
+    ncac_check_data_read_filter(&data_get, username);
 
     /* add output */
     node = lyd_new_output_anydata(output, NULL, "data", data_get, LYD_ANYDATA_DATATREE);
@@ -324,25 +294,29 @@ np2srv_rpc_get_cb(sr_session_ctx_t *session, const char *op_path, const struct l
     /* success */
 
 cleanup:
-    for (i = 0; i < filter_count; ++i) {
-        free(filters[i]);
-    }
-    free(filters);
+    op_filter_erase(&filter);
     lyd_free_withsiblings(data_get);
+    free(username);
     return rc;
 }
 
 int
 np2srv_rpc_editconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     sr_datastore_t ds = 0;
     struct ly_set *nodeset;
     struct lyd_node *config = NULL;
     const sr_error_info_t *err_info;
+    sr_session_ctx_t *user_sess = NULL;
     const char *str, *defop = "merge", *testop = "test-then-set";
     int rc = SR_ERR_OK;
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
     /* get know which datastore is being affected */
     nodeset = lyd_find_path(input, "target/*");
     if (!strcmp(nodeset->set.d[0]->schema->name, "running")) {
@@ -407,25 +381,33 @@ np2srv_rpc_editconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path),
     }
     ly_set_free(nodeset);
 
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, ds);
+    sr_session_switch_ds(user_sess, ds);
 
     /* sysrepo API */
     if (config) {
-        rc = sr_edit_batch(session, config, defop);
+        rc = sr_edit_batch(user_sess, config, defop);
         if (rc != SR_ERR_OK) {
             goto cleanup;
         }
     }
 
     if (!strcmp(testop, "test-then-set")) {
-        rc = sr_apply_changes(session, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+        rc = sr_apply_changes(user_sess, np2srv.sr_timeout, 1);
     } else {
         assert(!strcmp(testop, "test-only"));
-        rc = sr_validate(session, NULL, 0);
+        rc = sr_validate(user_sess, NULL, 0);
     }
     if (rc != SR_ERR_OK) {
-        sr_get_error(session, &err_info);
+        sr_get_error(user_sess, &err_info);
         sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
         goto cleanup;
     }
@@ -433,25 +415,52 @@ np2srv_rpc_editconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path),
     /* success */
 
 cleanup:
+    if (user_sess) {
+        /* discard any changes that possibly failed to be applied */
+        sr_discard_changes(user_sess);
+    }
     lyd_free_withsiblings(config);
     return rc;
 }
 
 int
 np2srv_rpc_copyconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     sr_datastore_t ds = SR_DS_OPERATIONAL, sds = SR_DS_OPERATIONAL;
     struct ly_set *nodeset;
     const sr_error_info_t *err_info;
     struct lyd_node *config = NULL;
     int rc = SR_ERR_OK, run_to_start = 0;
+    sr_session_ctx_t *user_sess;
+    char *username = NULL;
 #ifdef NP2SRV_URL_CAPAB
     struct lyd_node_leaf_list *leaf;
     const char *trg_url = NULL;
     int lyp_wd_flag;
 #endif
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
+    /* Get username. It is assumed that right now the NETCONF session cannot end
+     * due to the RPC lock held while np2srv_rpc_cb() is executing (which called this callback).
+     */
+    if ((username = (char *)np_get_nc_sess_user(session))) {
+        if (!(username = strdup(username))) {
+            EMEM;
+            rc = SR_ERR_NOMEM;
+            goto cleanup;
+        }
+    } else {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
+
     /* get know which datastores are affected */
     nodeset = lyd_find_path(input, "target/*");
     if (!strcmp(nodeset->set.d[0]->schema->name, "running")) {
@@ -528,11 +537,20 @@ np2srv_rpc_copyconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path),
         if (rc != SR_ERR_OK) {
             goto cleanup;
         }
-        ncac_check_data_read_filter(&config, np_get_nc_sess_user(session));
+
+        ncac_check_data_read_filter(&config, username);
+    }
+
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
     }
 
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, ds);
+    sr_session_switch_ds(user_sess, ds);
 
     /* sysrepo API/URL handling */
 #ifdef NP2SRV_URL_CAPAB
@@ -563,18 +581,18 @@ np2srv_rpc_copyconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path),
     {
         if (config) {
             /* config is spent */
-            rc = sr_replace_config(session, NULL, config, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+            rc = sr_replace_config(user_sess, NULL, config, np2srv.sr_timeout, 1);
             config = NULL;
         } else {
             assert(run_to_start);
 
             /* set SID to skip NACM check, only one copy-config can be executed at once */
-            ATOMIC_STORE_RELAXED(skip_nacm_sr_sid, sr_session_get_id(session));
-            rc = sr_copy_config(session, NULL, sds, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+            ATOMIC_STORE_RELAXED(skip_nacm_sr_sid, sr_session_get_id(user_sess));
+            rc = sr_copy_config(user_sess, NULL, sds, np2srv.sr_timeout, 1);
             ATOMIC_STORE_RELAXED(skip_nacm_sr_sid, 0);
         }
         if (rc != SR_ERR_OK) {
-            sr_get_error(session, &err_info);
+            sr_get_error(user_sess, &err_info);
             sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
             goto cleanup;
         }
@@ -584,21 +602,29 @@ np2srv_rpc_copyconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path),
 
 cleanup:
     lyd_free_withsiblings(config);
+    free(username);
     return rc;
 }
 
 int
 np2srv_rpc_deleteconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     sr_datastore_t ds = 0;
     struct ly_set *nodeset;
     int rc = SR_ERR_OK;
+    sr_session_ctx_t *user_sess;
+    const sr_error_info_t *err_info;
 #ifdef NP2SRV_URL_CAPAB
     struct lyd_node *config;
     const char *trg_url = NULL;
 #endif
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
     /* get know which datastore is affected */
     nodeset = lyd_find_path(input, "target/*");
     if (!strcmp(nodeset->set.d[0]->schema->name, "startup")) {
@@ -616,8 +642,16 @@ np2srv_rpc_deleteconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path
     }
     ly_set_free(nodeset);
 
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, ds);
+    sr_session_switch_ds(user_sess, ds);
 
     /* sysrepo API/URL handling */
 #ifdef NP2SRV_URL_CAPAB
@@ -636,8 +670,10 @@ np2srv_rpc_deleteconfig_cb(sr_session_ctx_t *session, const char *UNUSED(op_path
     } else
 #endif
     {
-        rc = sr_replace_config(session, NULL, NULL, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+        rc = sr_replace_config(user_sess, NULL, NULL, np2srv.sr_timeout, 1);
         if (rc != SR_ERR_OK) {
+            sr_get_error(user_sess, &err_info);
+            sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
             goto cleanup;
         }
     }
@@ -650,13 +686,19 @@ cleanup:
 
 int
 np2srv_rpc_un_lock_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     sr_datastore_t ds = 0;
     struct ly_set *nodeset;
+    sr_session_ctx_t *user_sess;
     const sr_error_info_t *err_info;
     int rc = SR_ERR_OK;
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
     /* get know which datastore is being affected */
     nodeset = lyd_find_path(input, "target/*");
     if (!strcmp(nodeset->set.d[0]->schema->name, "running")) {
@@ -669,17 +711,25 @@ np2srv_rpc_un_lock_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), co
     }
     ly_set_free(nodeset);
 
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, ds);
+    sr_session_switch_ds(user_sess, ds);
 
     /* sysrepo API */
     if (!strcmp(input->schema->name, "lock")) {
-        rc = sr_lock(session, NULL);
+        rc = sr_lock(user_sess, NULL);
     } else if (!strcmp(input->schema->name, "unlock")) {
-        rc = sr_unlock(session, NULL);
+        rc = sr_unlock(user_sess, NULL);
     }
     if (rc != SR_ERR_OK) {
-        sr_get_error(session, &err_info);
+        sr_get_error(user_sess, &err_info);
         sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
         goto cleanup;
     }
@@ -692,18 +742,23 @@ cleanup:
 
 int
 np2srv_rpc_kill_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     struct nc_session *kill_sess;
     struct ly_set *nodeset;
     uint32_t kill_sid, i;
     int rc = SR_ERR_OK;
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
     nodeset = lyd_find_path(input, "session-id");
     kill_sid = ((struct lyd_node_leaf_list *)nodeset->set.d[0])->value.uint32;
     ly_set_free(nodeset);
 
-    if (kill_sid == sr_session_get_nc_id(session)) {
+    if (kill_sid == sr_session_get_event_nc_id(session)) {
         rc = SR_ERR_INVAL_ARG;
         sr_set_error(session, NULL, "It is forbidden to kill own session.");
         goto cleanup;
@@ -723,7 +778,7 @@ np2srv_rpc_kill_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const
     /* kill the session */
     nc_session_set_status(kill_sess, NC_STATUS_INVALID);
     nc_session_set_term_reason(kill_sess, NC_SESSION_TERM_KILLED);
-    nc_session_set_killed_by(kill_sess, sr_session_get_nc_id(session));
+    nc_session_set_killed_by(kill_sess, kill_sid);
 
     /* success */
 
@@ -733,18 +788,32 @@ cleanup:
 
 int
 np2srv_rpc_commit_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *UNUSED(input),
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
-    int rc = SR_ERR_OK;;
+    int rc = SR_ERR_OK;
+    sr_session_ctx_t *user_sess;
     const sr_error_info_t *err_info;
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, SR_DS_RUNNING);
+    sr_session_switch_ds(user_sess, SR_DS_RUNNING);
 
     /* sysrepo API */
-    rc = sr_copy_config(session, NULL, SR_DS_CANDIDATE, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+    rc = sr_copy_config(user_sess, NULL, SR_DS_CANDIDATE, np2srv.sr_timeout, 1);
     if (rc != SR_ERR_OK) {
-        sr_get_error(session, &err_info);
+        sr_get_error(user_sess, &err_info);
         sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
         goto cleanup;
     }
@@ -757,16 +826,33 @@ cleanup:
 
 int
 np2srv_rpc_discard_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *UNUSED(input),
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     int rc = SR_ERR_OK;
+    sr_session_ctx_t *user_sess;
+    const sr_error_info_t *err_info;
+
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
 
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, SR_DS_CANDIDATE);
+    sr_session_switch_ds(user_sess, SR_DS_CANDIDATE);
 
     /* sysrepo API */
-    rc = sr_copy_config(session, NULL, SR_DS_RUNNING, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+    rc = sr_copy_config(user_sess, NULL, SR_DS_RUNNING, np2srv.sr_timeout, 1);
     if (rc != SR_ERR_OK) {
+        sr_get_error(user_sess, &err_info);
+        sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
         goto cleanup;
     }
 
@@ -778,12 +864,19 @@ cleanup:
 
 int
 np2srv_rpc_validate_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     sr_datastore_t ds;
     struct lyd_node *config = NULL;
     struct ly_set *nodeset;
+    sr_session_ctx_t *user_sess;
     int rc = SR_ERR_OK;
+    const sr_error_info_t *err_info;
+
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
 
     /* get know which datastore is affected */
     nodeset = lyd_find_path(input, "source/*");
@@ -821,11 +914,22 @@ np2srv_rpc_validate_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), c
     ly_set_free(nodeset);
 
     if (!config) {
-        /* update sysrepo session datastore */
-        sr_session_switch_ds(session, ds);
+        /* get the user session */
+        user_sess = np_get_user_sess(session);
+        if (!user_sess) {
+            EINT;
+            rc = SR_ERR_INTERNAL;
+            goto cleanup;
+        }
 
-        rc = sr_validate(session, NULL, 0);
+        /* update sysrepo session datastore */
+        sr_session_switch_ds(user_sess, ds);
+
+        /* sysrepo API */
+        rc = sr_validate(user_sess, NULL, 0);
         if (rc != SR_ERR_OK) {
+            sr_get_error(user_sess, &err_info);
+            sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
             goto cleanup;
         }
     }
@@ -837,64 +941,130 @@ cleanup:
     return rc;
 }
 
+static int
+np2srv_rpc_subscribe_append_str(const char *str, char **ret)
+{
+    void *mem;
+    int len;
+
+    if (!*ret) {
+        *ret = strdup(str);
+        if (!*ret) {
+            EMEM;
+            return SR_ERR_NOMEM;
+        }
+    } else {
+        len = strlen(*ret);
+        mem = realloc(*ret, len + strlen(str) + 1);
+        if (!mem) {
+            EMEM;
+            return SR_ERR_NOMEM;
+        }
+        *ret = mem;
+        strcat(*ret + len, str);
+    }
+
+    return SR_ERR_OK;
+}
+
 int
 np2srv_rpc_subscribe_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     struct ly_set *nodeset;
     const struct lys_module *ly_mod;
     struct lys_node *root, *next, *elem, *parent;
     struct nc_session *ncs;
     const char *stream;
-    char **filters = NULL, *xp = NULL, *mem;
+    struct np2_filter filter = {0};
+    char *xp = NULL;
     time_t start = 0, stop = 0;
-    int rc = SR_ERR_OK, i, len, filter_count = 0;
+    int rc = SR_ERR_OK, i;
     uint32_t idx;
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case (not supported) */
+        return SR_ERR_OK;
+    }
+
     /* find this NETCONF session */
     for (i = 0; (ncs = nc_ps_get_session(np2srv.nc_ps, i)); ++i) {
-        if (nc_session_get_id(ncs) == sr_session_get_nc_id(session)) {
+        if (nc_session_get_id(ncs) == sr_session_get_event_nc_id(session)) {
             break;
         }
     }
     if (!ncs) {
-        ERR("Failed to find NETCONF session SID %u.", sr_session_get_nc_id(session));
+        ERR("Failed to find NETCONF session SID %u.", sr_session_get_event_nc_id(session));
         rc = SR_ERR_INTERNAL;
         goto cleanup;
     }
 
+    /* RFC 5277 section 6.5 */
+    if (nc_session_get_notif_status(ncs)) {
+        sr_set_error(session, NULL, "Session already subscribed.");
+        return SR_ERR_EXISTS;
+    }
+
     /* learn stream */
     nodeset = lyd_find_path(input, "stream");
     stream = ((struct lyd_node_leaf_list *)nodeset->set.d[0])->value_str;
     ly_set_free(nodeset);
 
-    /* filter */
+    /* filter, join all into one xpath */
     nodeset = lyd_find_path(input, "filter");
     if (nodeset->number) {
-        if (op_filter_create(nodeset->set.d[0], &filters, &filter_count)) {
+        if (op_filter_create(nodeset->set.d[0], &filter)) {
             rc = SR_ERR_INTERNAL;
             goto cleanup;
         }
 
-        /* join all filters into one xpath */
-        for (i = 0; i < filter_count; ++i) {
+        /* all selection filters first */
+        for (i = 0; i < filter.count; ++i) {
+            if (!filter.filters[i].selection) {
+                continue;
+            }
+
+            /* put all selection filters into parentheses */
             if (!xp) {
-                xp = strdup(filters[0]);
-                if (!xp) {
-                    EMEM;
-                    rc = SR_ERR_NOMEM;
+                if ((rc = np2srv_rpc_subscribe_append_str("(", &xp))) {
+                    goto cleanup;
+                }
+
+                if ((rc = np2srv_rpc_subscribe_append_str(filter.filters[i].str, &xp))) {
                     goto cleanup;
                 }
             } else {
-                len = strlen(xp);
-                mem = realloc(xp, len + 5 + strlen(filters[i]) + 1);
-                if (!mem) {
-                    EMEM;
-                    rc = SR_ERR_NOMEM;
+                if ((rc = np2srv_rpc_subscribe_append_str(" or ", &xp))) {
                     goto cleanup;
                 }
-                xp = mem;
-                sprintf(xp + len, " and %s", filters[i]);
+
+                if ((rc = np2srv_rpc_subscribe_append_str(filter.filters[i].str, &xp))) {
+                    goto cleanup;
+                }
+            }
+        }
+
+        if (xp) {
+            /* finish parentheses */
+            if ((rc = np2srv_rpc_subscribe_append_str(")", &xp))) {
+                goto cleanup;
+            }
+        }
+
+        /* now append all content filters */
+        for (i = 0; i < filter.count; ++i) {
+            if (filter.filters[i].selection) {
+                continue;
+            }
+
+            if (xp) {
+                if ((rc = np2srv_rpc_subscribe_append_str(" and ", &xp))) {
+                    goto cleanup;
+                }
+            }
+
+            if ((rc = np2srv_rpc_subscribe_append_str(filter.filters[i].str, &xp))) {
+                goto cleanup;
             }
         }
     }
@@ -958,10 +1128,7 @@ np2srv_rpc_subscribe_cb(sr_session_ctx_t *session, const char *UNUSED(op_path),
     /* success */
 
 cleanup:
-    for (i = 0; i < filter_count; ++i) {
-        free(filters[i]);
-    }
-    free(filters);
+    op_filter_erase(&filter);
     free(xp);
     if (ncs && rc) {
         nc_session_set_notif_status(ncs, 0);
