@@ -726,11 +726,14 @@ ncac_rule_cb(sr_session_ctx_t *session, const char *UNUSED(module_name), const c
 }
 
 enum ncac_access {
-    NCAC_ACCESS_DENY = 1,       /**< access to the node is denied */
-    NCAC_ACCESS_PARTIAL = 2,    /**< access to the node is denied but it is a prefix of a matching rule */
-    NCAC_ACCESS_PERMIT = 3      /**< access to the node is permitted */
+    NCAC_ACCESS_DENY = 1,           /**< access to the node is denied */
+    NCAC_ACCESS_PARTIAL_DENY = 2,   /**< access to the node is denied but it is a prefix of a matching rule */
+    NCAC_ACCESS_PARTIAL_PERMIT = 3, /**< access to the node is permitted but any children must still be checked */
+    NCAC_ACCESS_PERMIT = 4          /**< access to the node is permitted with any children */
 };
 
+#define NCAC_ACCESS_IS_NODE_PERMIT(x) ((x) > 2)
+
 void
 ncac_init(void)
 {
@@ -786,6 +789,8 @@ ncac_getpwnam(const char *user, uid_t *uid, gid_t *gid)
     ssize_t buflen;
     int ret;
 
+    assert(user);
+
     buflen = sysconf(_SC_GETPW_R_SIZE_MAX);
     if (buflen == -1) {
         buflen = 2048;
@@ -985,9 +990,11 @@ cleanup:
  *
  * @param[in] rule_target Rule target instance-identifier.
  * @param[in] node_path Node data path.
- * @return NCAC access enum.
+ * @return 0 if does not match.
+ * @return 1 if the rule path matches.
+ * @return 2 if the path is a partial match.
  */
-static enum ncac_access
+static int
 ncac_allowed_path(const char *rule_target, const char *node_path)
 {
     const char *rule_ptr, *node_ptr;
@@ -1014,21 +1021,21 @@ ncac_allowed_path(const char *rule_target, const char *node_path)
             ++node_ptr;
         } else {
             /* not a match */
-            return NCAC_ACCESS_DENY;
+            return 0;
         }
     }
 
     if (!rule_ptr[0] && !node_ptr[0]) {
         /* full match */
-        return NCAC_ACCESS_PERMIT;
+        return 1;
     } else if (rule_ptr[0]) {
         assert(!node_ptr[0]);
         /* rule continues, it is a partial match */
-        return NCAC_ACCESS_PARTIAL;
+        return 2;
     } else {
         assert(!rule_ptr[0]);
         /* node continues, prefix (descendant) match */
-        return NCAC_ACCESS_PERMIT;
+        return 1;
     }
 }
 
@@ -1048,7 +1055,8 @@ ncac_allowed_node(const struct lyd_node *node, const char *user, uint8_t oper)
     struct ly_ctx *ly_ctx;
     char **groups, *path;
     uint32_t i, j, group_count;
-    enum ncac_access node_access, access = NCAC_ACCESS_DENY;
+    enum ncac_access access = NCAC_ACCESS_DENY, partial_access = NCAC_ACCESS_DENY;
+    int path_match;
 
     assert(oper);
 
@@ -1098,7 +1106,10 @@ ncac_allowed_node(const struct lyd_node *node, const char *user, uint8_t oper)
                 continue;
             }
 
-            node_access = NCAC_ACCESS_PERMIT;
+            /* access operation matching */
+            if (!(rule->operations & oper)) {
+                continue;
+            }
 
             /* target (rule) type matching */
             switch (rule->target_type) {
@@ -1130,24 +1141,22 @@ ncac_allowed_node(const struct lyd_node *node, const char *user, uint8_t oper)
                 if (rule->target) {
                     path = lyd_path(node);
                     /* exact match or is a descendant (specified in RFC 8341 page 27) for full tree access */
-                    node_access = ncac_allowed_path(rule->target, path);
+                    path_match = ncac_allowed_path(rule->target, path);
                     free(path);
-                    if (node_access == NCAC_ACCESS_DENY) {
+                    if (!path_match) {
+                        continue;
+                    } else if (path_match == 2) {
+                        /* partial match, continue searching for a full match (only partial permit has some meaning,
+                         * do not overwrite it) */
+                        partial_access = rule->action_deny ? partial_access : NCAC_ACCESS_PARTIAL_PERMIT;
                         continue;
                     }
                 }
                 break;
             }
 
-            /* access operation matching */
-            if (!(rule->operations & oper)) {
-                continue;
-            }
-
             /* 8) rule matched */
-            if (!rule->action_deny) {
-                access = node_access;
-            }
+            access = rule->action_deny ? NCAC_ACCESS_DENY : NCAC_ACCESS_PERMIT;
             goto cleanup;
         }
     }
@@ -1174,19 +1183,28 @@ step10:
     switch (oper) {
     case NCAC_OP_READ:
         if (nacm.default_read_deny) {
-            goto cleanup;
+            access = NCAC_ACCESS_DENY;
+        } else {
+            /* permit, but not by an explicit rule */
+            access = NCAC_ACCESS_PARTIAL_PERMIT;
         }
         break;
     case NCAC_OP_CREATE:
     case NCAC_OP_UPDATE:
     case NCAC_OP_DELETE:
         if (nacm.default_write_deny) {
-            goto cleanup;
+            access = NCAC_ACCESS_DENY;
+        } else {
+            /* permit, but not by an explicit rule */
+            access = NCAC_ACCESS_PARTIAL_PERMIT;
         }
         break;
     case NCAC_OP_EXEC:
         if (nacm.default_exec_deny) {
-            goto cleanup;
+            access = NCAC_ACCESS_DENY;
+        } else {
+            /* permit, but not by an explicit rule */
+            access = NCAC_ACCESS_PARTIAL_PERMIT;
         }
         break;
     default:
@@ -1194,8 +1212,10 @@ step10:
         goto cleanup;
     }
 
-    /* success */
-    access = NCAC_ACCESS_PERMIT;
+    if ((access == NCAC_ACCESS_DENY) && (partial_access == NCAC_ACCESS_PARTIAL_PERMIT)) {
+        /* node itself is not allowed but a rule allows access to some descendants so it may be allowed at the end */
+        access = NCAC_ACCESS_PARTIAL_DENY;
+    }
 
 cleanup:
     for (i = 0; i < group_count; ++i) {
@@ -1257,21 +1277,21 @@ ncac_check_operation(const struct lyd_node *data, const char *user)
 
     if (op->schema->nodetype & (LYS_RPC | LYS_ACTION)) {
         /* check X access on the RPC/action */
-        if (ncac_allowed_node(op, user, NCAC_OP_EXEC) != NCAC_ACCESS_PERMIT) {
+        if (!NCAC_ACCESS_IS_NODE_PERMIT(ncac_allowed_node(op, user, NCAC_OP_EXEC))) {
             goto cleanup;
         }
     } else {
         assert(op->schema->nodetype == LYS_NOTIF);
 
         /* check R access on the notification */
-        if (ncac_allowed_node(op, user, NCAC_OP_READ) != NCAC_ACCESS_PERMIT) {
+        if (!NCAC_ACCESS_IS_NODE_PERMIT(ncac_allowed_node(op, user, NCAC_OP_READ))) {
             goto cleanup;
         }
     }
 
     if (op->parent) {
         /* check R access on the parents, the last parent must be enough */
-        if (ncac_allowed_node(op->parent, user, NCAC_OP_READ) != NCAC_ACCESS_PERMIT) {
+        if (!NCAC_ACCESS_IS_NODE_PERMIT(ncac_allowed_node(op->parent, user, NCAC_OP_READ))) {
             goto cleanup;
         }
     }
@@ -1309,8 +1329,8 @@ ncac_check_data_read_filter_r(struct lyd_node **first, const char *user)
         /* check access of the node */
         node_access = ncac_allowed_node(elem, user, NCAC_OP_READ);
 
-        if (node_access == NCAC_ACCESS_PARTIAL) {
-            /* only partial access granted, we must check children recursively */
+        if (node_access == NCAC_ACCESS_PARTIAL_DENY) {
+            /* only partial deny access, we must check children recursively to learn whether this node is allowed or not */
             if (!(elem->schema->nodetype & (LYS_LEAF | LYS_LEAFLIST | LYS_ANYDATA))) {
                 node_access = ncac_check_data_read_filter_r(&elem->child, user);
             }
@@ -1319,6 +1339,12 @@ ncac_check_data_read_filter_r(struct lyd_node **first, const char *user)
                 /* none of the descendants are actually permitted, access denied */
                 node_access = NCAC_ACCESS_DENY;
             }
+        } else if (node_access == NCAC_ACCESS_PARTIAL_PERMIT) {
+            /* partial permit, the node will be included in the reply but we must check children as well */
+            if (!(elem->schema->nodetype & (LYS_LEAF | LYS_LEAFLIST | LYS_ANYDATA))) {
+                ncac_check_data_read_filter_r(&elem->child, user);
+            }
+            node_access = NCAC_ACCESS_PERMIT;
         }
 
         /* access denied, free the subtree */
@@ -1410,7 +1436,7 @@ ncac_check_diff_r(const struct lyd_node *diff, const char *user, const char *par
         }
 
         /* check access for the node, none operation is always allowed, and partial access is relevant only for read operation */
-        if (oper && (ncac_allowed_node(diff, user, oper) != NCAC_ACCESS_PERMIT)) {
+        if (oper && !NCAC_ACCESS_IS_NODE_PERMIT(ncac_allowed_node(diff, user, oper))) {
             node = diff;
             break;
         }
