@@ -90,17 +90,39 @@ op_data_filter_origin(struct lyd_node **data, const struct lys_ident *filter, in
 
 int
 np2srv_rpc_getdata_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *output, void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *output, void *UNUSED(private_data))
 {
     struct lyd_node_leaf_list *leaf;
-    struct lyd_node *node, *data_get = NULL;
-    char **filters = NULL;
-    int filter_count = 0, i, rc = SR_ERR_OK;
+    struct lyd_node *node, *select_data = NULL, *data = NULL;
+    struct np2_filter filter = {0};
+    int i, rc = SR_ERR_OK;
+    sr_session_ctx_t *user_sess;
     uint32_t max_depth = 0;
     struct ly_set *nodeset;
     sr_datastore_t ds;
     NC_WD_MODE nc_wd;
     sr_get_oper_options_t get_opts = 0;
+    char *username = NULL;
+
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
+    /* Get username. It is assumed that right now the NETCONF session cannot end
+     * due to the RPC lock held while np2srv_rpc_cb() is executing (which called this callback).
+     */
+    if ((username = (char *)np_get_nc_sess_user(session))) {
+        if (!(username = strdup(username))) {
+            EMEM;
+            rc = SR_ERR_NOMEM;
+            goto cleanup;
+        }
+    } else {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
 
     /* get default value for with-defaults */
     nc_server_get_capab_withdefaults(&nc_wd, NULL);
@@ -128,24 +150,25 @@ np2srv_rpc_getdata_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), co
     node = nodeset->number ? nodeset->set.d[0] : NULL;
     ly_set_free(nodeset);
     if (node && !strcmp(node->schema->name, "subtree-filter")) {
-        if (op_filter_create(node, &filters, &filter_count)) {
+        if (op_filter_create(node, &filter)) {
             rc = SR_ERR_INTERNAL;
             goto cleanup;
         }
     } else {
-        filters = malloc(sizeof *filters);
-        if (!filters) {
+        filter.filters = malloc(sizeof *filter.filters);
+        if (!filter.filters) {
             EMEM;
             rc = SR_ERR_NOMEM;
             goto cleanup;
         }
-        filter_count = 1;
-        filters[0] = node ? strdup(((struct lyd_node_leaf_list *)node)->value_str) : strdup("/*");
-        if (!filters[0]) {
+        filter.count = 1;
+        filter.filters[0].str = node ? strdup(((struct lyd_node_leaf_list *)node)->value_str) : strdup("/*");
+        if (!filter.filters[0].str) {
             EMEM;
             rc = SR_ERR_NOMEM;
             goto cleanup;
         }
+        filter.filters[0].selection = 1;
     }
 
     /* config filter */
@@ -193,69 +216,73 @@ np2srv_rpc_getdata_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), co
         }
     }
 
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, ds);
+    sr_session_switch_ds(user_sess, ds);
 
     /*
      * create the data tree for the data reply
      */
-    for (i = 0; i < filter_count; i++) {
-        rc = sr_get_data(session, filters[i], max_depth, np2srv.sr_timeout, get_opts, &node);
-        if (rc != SR_ERR_OK) {
-            ERR("Getting data \"%s\" from sysrepo failed (%s).", filters[i], sr_strerror(rc));
-            goto cleanup;
-        }
-
-        if (!data_get) {
-            data_get = node;
-        } else if (node && lyd_merge(data_get, node, LYD_OPT_DESTRUCT | LYD_OPT_EXPLICIT)) {
-            lyd_free_withsiblings(node);
-            rc = SR_ERR_LY;
-            goto cleanup;
-        }
+    if ((rc = op_filter_data_get(user_sess, max_depth, get_opts, &filter, session, &select_data))) {
+        goto cleanup;
+    }
+    if ((rc = op_filter_data_filter(&select_data, &filter, 0, &data))) {
+        goto cleanup;
     }
 
     /* origin filter */
     nodeset = lyd_find_path(input, "origin-filter | negated-origin-filter");
     for (i = 0; i < (signed)nodeset->number; ++i) {
         leaf = (struct lyd_node_leaf_list *)nodeset->set.d[i];
-        op_data_filter_origin(&data_get, leaf->value.ident, strcmp(leaf->schema->name, "origin-filter"));
+        op_data_filter_origin(&data, leaf->value.ident, strcmp(leaf->schema->name, "origin-filter"));
     }
     ly_set_free(nodeset);
 
     /* perform correct NACM filtering */
-    ncac_check_data_read_filter(&data_get, np_get_nc_sess_user(session));
+    ncac_check_data_read_filter(&data, username);
 
     /* add output */
-    node = lyd_new_output_anydata(output, NULL, "data", data_get, LYD_ANYDATA_DATATREE);
+    node = lyd_new_output_anydata(output, NULL, "data", data, LYD_ANYDATA_DATATREE);
     if (!node) {
         goto cleanup;
     }
-    data_get = NULL;
+    data = NULL;
 
     /* success */
 
 cleanup:
-    for (i = 0; i < filter_count; ++i) {
-        free(filters[i]);
-    }
-    free(filters);
-    lyd_free_withsiblings(data_get);
+    op_filter_erase(&filter);
+    lyd_free_withsiblings(select_data);
+    lyd_free_withsiblings(data);
+    free(username);
     return rc;
 }
 
 int
 np2srv_rpc_editdata_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), const struct lyd_node *input,
-        sr_event_t UNUSED(event), uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
+        sr_event_t event, uint32_t UNUSED(request_id), struct lyd_node *UNUSED(output), void *UNUSED(private_data))
 {
     sr_datastore_t ds;
     struct ly_set *nodeset;
     struct lyd_node_leaf_list *leaf;
     struct lyd_node *node, *config = NULL;
     const sr_error_info_t *err_info;
+    sr_session_ctx_t *user_sess;
     const char *defop;
     int rc = SR_ERR_OK;
 
+    if (event == SR_EV_ABORT) {
+        /* ignore in this case */
+        return SR_ERR_OK;
+    }
+
     /* get know which datastore is being affected */
     nodeset = lyd_find_path(input, "datastore");
     leaf = (struct lyd_node_leaf_list *)nodeset->set.d[0];
@@ -301,18 +328,26 @@ np2srv_rpc_editdata_cb(sr_session_ctx_t *session, const char *UNUSED(op_path), c
 #endif
     }
 
+    /* get the user session */
+    user_sess = np_get_user_sess(session);
+    if (!user_sess) {
+        EINT;
+        rc = SR_ERR_INTERNAL;
+        goto cleanup;
+    }
+
     /* update sysrepo session datastore */
-    sr_session_switch_ds(session, ds);
+    sr_session_switch_ds(user_sess, ds);
 
     /* sysrepo API */
-    rc = sr_edit_batch(session, config, defop);
+    rc = sr_edit_batch(user_sess, config, defop);
     if (rc != SR_ERR_OK) {
         goto cleanup;
     }
 
-    rc = sr_apply_changes(session, np2srv.sr_timeout, NP2SRV_DATA_CHANGE_WAIT);
+    rc = sr_apply_changes(user_sess, np2srv.sr_timeout, 1);
     if (rc != SR_ERR_OK) {
-        sr_get_error(session, &err_info);
+        sr_get_error(user_sess, &err_info);
         sr_set_error(session, err_info->err[0].xpath, err_info->err[0].message);
         goto cleanup;
     }
